% This is an example demonstrating how to correctly extract data from block
% file format files.
% published on May 8, 2023
clc; clear;

%% ===============Set up for file parse============= %%

folder = 'C:\';
file = 'NEUR0001.DF1';
fileName = fullfile(folder, file);
fid = fopen(fileName, 'r');
rawData = uint8(fread(fid, Inf, 'uint8'));
fclose(fid);

%% ===================== User initialized neural values ===================== %%
numberOfNeuralAdcBits = 16; %of neural data
isNeuralSigned = true;
voltageResolution = 1.95e-7;
numberOfChannels = 64;
fsNeural = 32000; 
%% ===================== User initialized audio values ===================== %%
numberOfAudioBits = 15;
fsAudio = 200000;
isAudioSigned = true;

%% ================= User initialized motion sensor values ================= %%
% The values for acclMax and gyroMax are chosen by the user. They can be found using the Event
% File Viewer in the file started event.
acclMax = 2*MotionSensorConstants.G; % m/s^2, maximum value of selected range
gyroMax = 250; %  degrees per second,  maximum value of selected range
magMax = MotionSensorConstants.Magnetometer9250Range; %  Teslas,  maximum value of selected range
%% ================Extract metadata from block header =============== %%


constId = (hex2num(HeaderConstants.HexConstId));
constIdBytes = typecast(constId, 'uint8');
blockStartIndices = FindDataBlockStart(rawData, constIdBytes);
numberOfBlocks = length(blockStartIndices);
blockSize = length(rawData) / numberOfBlocks;
startOfFirstHeader = blockStartIndices(1);
endOfFirstHeader = startOfFirstHeader + HeaderConstants.HeaderTotalBytes;
firstHeader = rawData(startOfFirstHeader:endOfFirstHeader);
HeaderStruct = ExtractHeaderData(firstHeader);

%% ================Extract neural data from blocks =============== %%

%check where each type data is in partition info
neuralIndex = find(cell2mat(arrayfun(@(x) x.DataType == uint32(DataTypeEnum.NeuralData), HeaderStruct.PartitionInfo, 'un', 0)));
isNeuralPresent = ~isempty(neuralIndex);
if isNeuralPresent
    neuralDataAsBytes = ExtractDataByType(rawData, HeaderStruct, neuralIndex, blockStartIndices, numberOfBlocks);
end


%% ================Convert neural data to physical units (V) =============== %%
% get meta data from file start event using event file reader
if isNeuralPresent

    if isNeuralSigned %gets offset based on whether data are signed or not
        offset = 0;
    else
        offset = 2 ^ (numberOfNeuralAdcBits - 1);
    end
    %cast bytes to unsigned 16 bit integers and store as float
    neuralData= single(typecast(neuralDataAsBytes, 'uint16'));

    %convert uint16 values to voltage values
    for dataPointIndex = 1:length(neuralData)
        neuralData(dataPointIndex) = voltageResolution * (neuralData(dataPointIndex) - offset); 
    end

    % sort by channels
    neuralDataMat = reshape(neuralData, numberOfChannels, []);

    % get timestamps of neural data
    timestampsNeural = GetTimestamps(HeaderStruct.Timestamp, fsNeural, size(neuralDataMat, 2));
end
%% ================Extract audio data from blocks =============== %%

%check where each type data is in partition info
audioIndex = find(cell2mat(arrayfun(@(x) x.DataType == uint32(DataTypeEnum.Audio), HeaderStruct.PartitionInfo, 'un', 0)));
isAudioPresent = ~isempty(audioIndex);
if isAudioPresent
    audioDataAsBytes = ExtractDataByType(rawData, HeaderStruct, audioIndex, blockStartIndices, numberOfBlocks);
end

%% ================Scale audio data and save as wav =============== %%
% get meta data from file start event using event file reader

if isAudioPresent

    audioData = ScaleAudioData(isAudioSigned, audioDataAsBytes, numberOfAudioBits); 
    % save audio data as wav file
%     audiowrite('C:\Users\myPath\example200kHz.wav', audioData, fsAudio, 'BitsPerSample', 16);
    % get timestamps of audio data
    timestampsAudio = GetTimestamps(HeaderStruct.Timestamp, fsAudio, length(audioData));
end

%% ================Extract motion sensor data from blocks =============== %%

%check where each type data is in partition info
motionSensorIndex = find(cell2mat(arrayfun(@(x) x.DataType == uint32(DataTypeEnum.MotionSensor), HeaderStruct.PartitionInfo, 'un', 0)));
isMotionSensorPresent = ~isempty(motionSensorIndex);
if isMotionSensorPresent
    motionSensorAsBytes = ExtractDataByType(rawData, HeaderStruct, motionSensorIndex, blockStartIndices, numberOfBlocks);

end

%% ================Sort motion sensor data by data type =============== %%


% extract motion sensor data from inner block

if isMotionSensorPresent

    
    motionSensorData = single(typecast(motionSensorAsBytes, 'int16'));    
    blockStartIndicesMs = FindMotionSensorBlockStart(motionSensorData, MotionSensorConstants.ConstId);    
    accelerometerDataTemp = ExtractMotionSensorDataByType(motionSensorData, blockStartIndicesMs, MotionSensorEnum.Accelerometer);
    gyroscopeDataTemp = ExtractMotionSensorDataByType(motionSensorData, blockStartIndicesMs, MotionSensorEnum.Gyroscope);
    magnetometerDataTemp = ExtractMotionSensorDataByType(motionSensorData, blockStartIndicesMs, MotionSensorEnum.Magnetometer);
    
    scaledAccelerometer = ScaleMotionSensorData(accelerometerDataTemp, MotionSensorConstants.AccelerometerNumberOfBits, acclMax);
    scaledGyroscope = ScaleMotionSensorData(gyroscopeDataTemp, MotionSensorConstants.GyroscopeNumberOfBits, gyroMax);
    scaledMagnetometer = ScaleMotionSensorData(magnetometerDataTemp, MotionSensorConstants.Magnetometer9250NumberOfBits, magMax);

    AccelerometerData = SortDataByAxis(scaledAccelerometer);
    GyroscopeData = SortDataByAxis(scaledGyroscope);
    MagnetometerData = SortDataByAxis(scaledMagnetometer);
    
    % get timestamp of a particular point
    timestampsAccelerometer = GetMotionSensorTimestamp(motionSensorData, blockStartIndicesMs, MotionSensorEnum.Accelerometer, MotionSensorConstants.AccelerometerFrequency);
    timestampsGyroscope = GetMotionSensorTimestamp(motionSensorData, blockStartIndicesMs, MotionSensorEnum.Gyroscope, MotionSensorConstants.GyroscopeFrequency);
    timestampsMagnetometer = GetMotionSensorTimestamp(motionSensorData, blockStartIndicesMs, MotionSensorEnum.Magnetometer, MotionSensorConstants.MagnetometerFrequency);

    % example: plot accelerometer
    figure;
    plot(timestampsAccelerometer, AccelerometerData.X);
    hold on;
    plot(timestampsAccelerometer, AccelerometerData.Y);
    hold on;
    plot(timestampsAccelerometer, AccelerometerData.Z);
    xlabel('timestamp (ms from midnight)');
    ylabel('m/s^2');
    title('Accelerometer data');

end



